@@ -4,14 +4,18 @@
 build XML data in CMI format
 """
 
+import uuid
 from lxml import etree
 
-
+XML_NS = "http://www.w3.org/XML/1998/namespace"
 TEI_NS = "http://www.tei-c.org/ns/1.0"
 RNG_SCHEMA = "https://raw.githubusercontent.com/TEI-Correspondence-SIG/" + \
     "CMIF/master/schema/cmi-customization.rng"
 PI_TEXT = "href=\""+RNG_SCHEMA+"\" type=\"application/xml\" schematypens" + \
     "=\"http://relaxng.org/ns/structure/1.0\""
+CC_TEXT = "This file is licensed under the terms " \
+            + "of the Creative-Commons-License CC-BY 4.0"
+CC_URL = "https://creativecommons.org/licenses/by/4.0/"
 
 
 def pi_rng():
@@ -21,6 +25,13 @@ def pi_rng():
     return etree.ProcessingInstruction("xml-model", PI_TEXT)
 
 
+def ns_xml(attrib):
+    """
+    add xml namespace to given attribute
+    """
+    return "{" + XML_NS + "}" + attrib
+
+
 def tei_root(children=None):
     """
     create TEI root element <TEI> with (optional) children
@@ -137,9 +148,8 @@ def tei_license(elem_text="", attrib_target=""):
     create TEI element <licence> with (optional) text and @target
     """
     if elem_text == "" and attrib_target == "":
-        elem_text = "This file is licensed under the terms" \
-                        + " of the Creative-Commons-License CC-BY 4.0"
-        attrib_target = "https://creativecommons.org/licenses/by/4.0/"
+        elem_text = CC_TEXT
+        attrib_target = CC_URL
     license = etree.Element("licence")
     license.set("target", attrib_target)
     license.text = elem_text
@@ -148,20 +158,24 @@ def tei_license(elem_text="", attrib_target=""):
 
 def tei_source_desc(children=None):
     """
-    create TEI element <sourceDesc>
+    create TEI element <sourceDesc> with (optional) children
     """
     source_desc = etree.Element("sourceDesc")
     add_children(source_desc, children)
     return source_desc
 
 
-def tei_bibl(elem_text, attrib_type):
+def tei_bibl(elem_text, attrib_type, attrib_xml_id=None, domain=None):
     """
     create TEI element <bibl> with given text and @type
     """
     bibl = etree.Element("bibl")
     bibl.set("type", attrib_type)
     bibl.text = elem_text
+    if attrib_xml_id is None:
+        attrib_xml_id = str(uuid.uuid3(uuid.NAMESPACE_URL, domain)) if domain \
+            else str(uuid.uuid4())
+    bibl.set(ns_xml("id"), attrib_xml_id)
     return bibl
 
 
@@ -174,20 +188,22 @@ def tei_profile_desc(children=None):
     return profile_desc
 
 
-def tei_corresp_desc(attrib_ref="", children=None):
+def tei_corresp_desc(attrib_key="", attrib_ref="",
+                     attrib_source="", children=None):
     """
-    create TEI <correspDesc> element with @ref
+    create TEI element <correspDesc> with @ref
     """
     corresp_desc = etree.Element("correspDesc")
-    if attrib_ref != "":
-        corresp_desc.set("ref", attrib_ref)
+    add_attrib(corresp_desc, "key", attrib_key)
+    add_attrib(corresp_desc, "ref", attrib_ref)
+    add_attrib(corresp_desc, "source", attrib_source)
     add_children(corresp_desc, children)
     return corresp_desc
 
 
 def tei_corresp_action(attrib_type, children=None):
     """
-    create TEI <correspAction> element with @type
+    create TEI element <correspAction> with @type
     """
     corresp_action = etree.Element("correspAction")
     corresp_action.set("type", attrib_type)
@@ -195,39 +211,49 @@ def tei_corresp_action(attrib_type, children=None):
     return corresp_action
 
 
-def tei_date(attrib_when="", attrib_from="", attrib_to=""):
+def tei_date(attrib_when="", attrib_from="", attrib_to="",
+             attrib_not_before="", attrib_not_after=""):
     """
-    create TEI <date> element with @when or @from and @to
+    create TEI element <date> with @when or @from and @to
     """
     date = etree.Element("date")
-    if attrib_when != "":
-        date.set("when", attrib_when)
-    elif attrib_from != "" and attrib_to != "":
-        date.set("from", attrib_from)
-        date.set("to", attrib_to)
+    add_attrib(date, "when", attrib_when)
+    add_attrib(date, "from", attrib_from)
+    add_attrib(date, "to", attrib_to)
+    add_attrib(date, "notBefore", attrib_not_before)
+    add_attrib(date, "notAfter", attrib_not_after)
     return date
 
 
 def tei_place_name(elem_text, attrib_ref=""):
     """
-    create TEI <placeName> element with given element text and @ref
+    create TEI element <placeName> with given element text and @ref
     """
     place_name = etree.Element("placeName")
     place_name.text = elem_text
-    place_name.set("ref", attrib_ref)
+    add_attrib(place_name, "ref", attrib_ref)
     return place_name
 
 
 def tei_pers_name(elem_text, attrib_ref=""):
     """
-    create TEI <persName> element with given element text and @ref
+    create TEI element <persName> with given element text and @ref
     """
     pers_name = etree.Element("persName")
     pers_name.text = elem_text
-    pers_name.set("ref", attrib_ref)
+    add_attrib(pers_name, "ref", attrib_ref)
     return pers_name
 
 
+def tei_org_name(elem_text):
+    """
+    create TEI element <orgName> with given element text
+    """
+    org_name = etree.Element("orgName")
+    org_name.text = elem_text
+    return org_name
+
+
 def tei_text_empty():
     """
     create TEI element <text> with child elements <body> and <p> (empty)
@@ -241,7 +267,7 @@ def tei_text_empty():
 
 def tei_text():
     """
-    create TEI element <text> with child elements <body> and <p> (empty)
+    create TEI element <text>
     """
     return etree.Element("text")
 
@@ -267,6 +293,14 @@ def add_pi(tree):
     tree.getroot().addprevious(pi_rng())
 
 
+def add_attrib(element, name, value):
+    """
+    add attribute to element if value != ""
+    """
+    if value != "":
+        element.set(name, value)
+
+
 def add_child(parent, element):
     """
     add element to parent if not None
